import std.stdio;
import std.stream;
import std.string;

enum STONE{
  NON   = 0 ,
  BLACK = 1 ,
  WHITE = 2
} ;

enum PLAYRESULT{
  NON,
  WHITE_WIN,
  BLACK_WIN,
  DRAW
} ;

class Board {
  STONE matrix[][] ;
  STONE pre_matrix[][] ;
  int   w, h ;
  int chkvec_x[8] ;
  int chkvec_y[8] ;
  PLAYRESULT result ;

  this(){
    chkvec_x[0]= 1 ; chkvec_y[0]= 0 ;
    chkvec_x[1]=-1 ; chkvec_y[1]= 0 ;
    chkvec_x[2]= 0 ; chkvec_y[2]= 1 ;
    chkvec_x[3]= 0 ; chkvec_y[3]=-1 ;
    chkvec_x[4]= 1 ; chkvec_y[4]= 1 ;
    chkvec_x[5]=-1 ; chkvec_y[5]= 1 ;
    chkvec_x[6]= 1 ; chkvec_y[6]=-1 ;
    chkvec_x[7]=-1 ; chkvec_y[7]=-1 ;
  }

  int init(int w, int h){
    this.w=w ;
    this.h=h ;

    matrix.length=h ;
    for( int y=0 ; y<h ; y++ ){
      matrix[y].length=w ;
      matrix[y][]=STONE.NON ;
    }
    matrix[h/2-1][w/2-1]=STONE.BLACK ; matrix[h/2-1][w/2  ]=STONE.WHITE ;  // X,O
    matrix[h/2  ][w/2-1]=STONE.WHITE ; matrix[h/2  ][w/2  ]=STONE.BLACK ;  // O,X

    result=PLAYRESULT.NON ;
    
    return(0) ;
  }

  int setStone(int x, int y, STONE stone){
    if( checkCount(stone)>0 ){
      if( reverse(x,y,stone)==0 ) return(1) ;
      matrix[y][x]=stone ;
    }
    return(0) ;
  }

  int reverse(int x, int y, STONE stone){
    pre_matrix = getCroneBoard() ;
    
    if( x<0 || x>7 || y<0 || y>7 ) return(0) ;
    if( matrix[y][x]!=STONE.NON  ) return(0) ;
    
    int total=0 ;
    for( int chk=0 ; chk<8 ; chk++ ){
      int dx=chkvec_x[chk] ;
      int dy=chkvec_y[chk] ;
      int sx=x+dx ;
      int sy=y+dy ;
      int count=0 ;
      int stop =0 ;
      while(-1){
	if( sx>=0 && sy>=0 && sx<w && sy<h ){
	  if( matrix[sy][sx]!=STONE.NON && matrix[sy][sx]!=stone ){
	    count++ ;
	    sx+=dx ; sy+=dy ;
	    continue ;
	  }
	  if( matrix[sy][sx]==stone ) stop=1 ;
	}
	break ;
      }
      if( count>0 && stop==1 ){
	sx=x+dx ;
	sy=y+dy ;
	for( int i=0 ; i<count ; i++,sx+=dx,sy+=dy ){
	  matrix[sy][sx]=stone ;
	}
	total+=count ;
      }
    }
    return(total) ;
  }

  int Count(int x, int y, STONE stone){
    int total=0 ;
    
    for( int chk=0 ; chk<8 ; chk++ ){
      int dx=chkvec_x[chk] ;
      int dy=chkvec_y[chk] ;
      int sx=x+dx ;
      int sy=y+dy ;
      int count=0 ;
      int stop =0 ;
      while(-1){
	if( sx>=0 && sy>=0 && sx<w && sy<h ){
	  if( matrix[sy][sx]!=STONE.NON && matrix[sy][sx]!=stone ){
	    count++ ;
	    sx+=dx ; sy+=dy ;
	    continue ;
	  }
	  if( matrix[sy][sx]==stone ) stop=1 ;
	}
	break ;
      }
      if( count>0 && stop==1 ){
	total+=count ;
      }
    }
    return(total) ;
  }

  int checkCount(STONE stone){
    int cnt=0 ;
    for( int y=0 ; y<h ; y++ ){
      for( int x=0 ; x<w ; x++ ){
	if( matrix[y][x]==STONE.NON ){
	  cnt += Count(x,y,stone) ;
	}
      }
    }
    if( cnt>0 ) return(1) ;
    return(0) ;
  }
  
  bool isGameOver(STONE turn){
    bool is_over  = false ;
    bool w_ispass = checkCount(STONE.WHITE)==0 ? true : false ;
    bool b_ispass = checkCount(STONE.BLACK)==0 ? true : false ;
    if( turn==STONE.WHITE ){
      if( w_ispass==true && b_ispass==true ) is_over=true ;
    }else{
      if( b_ispass==true && w_ispass==true ) is_over=true ;
    }

    if( is_over==true ){
      int w_count=0 ;
      int b_count=0 ;
      int n_count=0 ;
      for( int y=0 ; y<h ; y++ ){
	for( int x=0 ; x<w ; x++ ){
	  switch(matrix[y][x]){
	  case STONE.BLACK :
	    b_count++ ; break ;
	  case STONE.WHITE :
	    w_count++ ; break ;
	  case STONE.NON :
	    n_count++ ; break ;
	  default : break ;
	  }
	}
      }
      result = (w_count==b_count) ? PLAYRESULT.DRAW : (w_count>b_count) ? PLAYRESULT.WHITE_WIN : PLAYRESULT.BLACK_WIN ;
      return(true) ;
    }
    return(false) ;
  }

  PLAYRESULT getResult(){
    return(result) ;
  }
  
  STONE[][] getCroneBoard(){
    STONE[][] mat ;

    mat.length=h ;
    for( int y=0 ; y<h ; y++ ){
      mat[y].length=w ;
      mat[y][]=matrix[y][] ;
    }
    return( mat ) ;
  }
  
  int printBoard(){
    int b_count=0 ;
    int w_count=0 ;
    
    writef("===\n") ;
    writef("+-+-+-+-+-+-+-+-+\n") ;
    for( int y=0 ; y<h ; y++ ){
      writef("|") ;
      for( int x=0 ; x<w ; x++ ){
	switch(matrix[y][x]){
	case STONE.WHITE :
	  writef("O") ; w_count++ ; break ;
	case STONE.BLACK :
	  writef("X") ; b_count++ ; break ;
	default :
	  writef(" ") ; break ;
	}
	writef("|") ;
      }
      writef("\n") ;
      writef("+-+-+-+-+-+-+-+-+\n") ;
    }
    writef("    O=%d  X=%d\n",w_count, b_count) ;
    return(0) ;
  }

}


version(DEBUG){

import Com ;

int main(string[] args)
{
  int x,y ;
  
  Board bd  = new Board() ; bd.init(8,8) ;
  Com   com1= new Com(STONE.BLACK) ;
  Com   com2= new Com(STONE.WHITE) ;
  
  com1.setRawBoard(bd) ;
  com2.setRawBoard(bd) ;

  bd.printBoard() ;
  for( int i=0 ; i<64 ; i++ ){
    com1.think() ;
    bd.setStone(com1.getX(), com1.getY(), com1.getStone()) ;
    bd.printBoard() ;
    if( bd.isGameOver()==true ) break ;
    com2.think() ;
    bd.setStone(com2.getX(), com2.getY(), com2.getStone()) ;
    bd.printBoard() ;
    if( bd.isGameOver()==true ) break ;
  }

  //  bd.setStone(3,5,STONE.BLACK) ;
  //  bd.setStone(2,5,STONE.WHITE) ;
  //  bd.setStone(2,4,STONE.BLACK) ;
  //  bd.setStone(4,5,STONE.WHITE) ;
  
  return(0) ;
}

}  