import win32.windows;
import win32frame.glut;
import std.stdio;
import core.thread;
import std.math;

import mt19937ar ;
import GameFrame;
import Sprite;
import SPformat;

import Vec3D;
import Quaternion;

import Board;
import Com;

enum GAMEST{
  TITLE,
  PLAYING,
  GAMEOVER
} ;

enum PLAYST{
  INIT,
  IDLE,
  PLAYER,
  REVANIM
} ;

enum DRAWST{
  IDLE,
  REVERSE
} ;

enum PLAYER{
  COM,
  YOU
} ;


class GameSequence
{
  Board  bd   ;
  Com    com1 ;
  GAMEST gamest ;

  double    g_rot_x ;
  int       wait   ;
  PLAYST    playst ;
  int       turn   ;
  PLAYER[2] player ;
  STONE     your_stone ;

  vec3D stonepos ;
  vec3D putpos   ;
  int   put_x, put_y ;
  int   msLbtn ;
  
  this(){
  }

  int init(Sprite sp, GameInfo* gi){
    bd  = new Board() ; bd.init(8,8) ;
    gamest = GAMEST.TITLE ;
    g_rot_x = 60.0 ;
    return(0) ;
  }

  int makeframe(Sprite sp, GameInfo* gi){
    int ret ;
    switch(gamest){
    case GAMEST.TITLE :
      ret=title(sp,gi) ;
      if( ret==1 ){
	gamest = GAMEST.PLAYING ;
	playst = PLAYST.INIT ;
      }
      break ;
    case GAMEST.PLAYING :
      ret=playing(sp,gi) ;
      if( ret==1 ) gamest=GAMEST.GAMEOVER ;
      break ;
    case GAMEST.GAMEOVER :
      ret=gameover(sp,gi) ;
      if( ret==1 ){
	bd.init(8,8) ;
	gamest=GAMEST.TITLE ;
      }
      break ;
    }
    return(0) ;
  }

  int title(Sprite sp, GameInfo* gi){
    int ret=0 ;
    sp.set2DSP(0,256,90+42,0.0,255,1.0) ;
    if( msLbtn!=0 && (gi.mouseb & MK_LBUTTON)==0 ){
      ret=1 ;
    }
    msLbtn = (gi.mouseb & MK_LBUTTON)!=0 ;
    setDrawPlayBD(sp,gi,DRAWST.IDLE) ;
    return(ret) ;
  }

  int gameover(Sprite sp, GameInfo* gi){
    int ret=0 ;
    switch(bd.getResult()){
    case PLAYRESULT.DRAW :
      sp.set2DSP(3,256,230,0.0,255,1.0) ;
      break ;
    case PLAYRESULT.BLACK_WIN :
      if( your_stone==STONE.BLACK ){
	sp.set2DSP(1,256,230,0.0,255,1.0) ;
      }else{
	sp.set2DSP(2,256,230,0.0,255,1.0) ;
      }
      break ;
    case PLAYRESULT.WHITE_WIN :
      if( your_stone==STONE.WHITE ){
	sp.set2DSP(1,256,230,0.0,255,1.0) ;
      }else{
	sp.set2DSP(2,256,230,0.0,255,1.0) ;
      }
      break ;
    }
    if( msLbtn!=0 && (gi.mouseb & MK_LBUTTON)==0 ){
      ret=1 ;
    }
    msLbtn = (gi.mouseb & MK_LBUTTON)!=0 ;
    setDrawPlayBD(sp,gi,DRAWST.IDLE) ;
    return(ret) ;
  }

  int playing(Sprite sp, GameInfo* gi){
    int ret=0 ;
    switch(playst){
    case PLAYST.INIT :
      bd.init(8,8) ;
      player[0] = PLAYER.YOU ;
      player[1] = PLAYER.COM ;
      your_stone= STONE.BLACK;
      com1= new Com(STONE.WHITE) ;
      com1.setRawBoard(bd) ;
      turn   = 0 ;
      playst = PLAYST.IDLE ;
      break ;
    case PLAYST.IDLE :
      wait++ ;
      if( wait>10 ){
	wait=0 ;
	playst = PLAYST.PLAYER ;
      }
      break ;
    case PLAYST.PLAYER :
      STONE turnstone = (player[turn]==PLAYER.YOU) ? your_stone : your_stone==STONE.BLACK ? STONE.WHITE : STONE.BLACK ;
      if( bd.isGameOver(turnstone)!=true ){
	switch( player[turn] ){
	case PLAYER.COM :
	  com1.think() ;
	  bd.setStone(com1.getX(), com1.getY(), com1.getStone()) ;
	  turn=(turn+1)%2 ;
	  playst=PLAYST.REVANIM ;
	  break ;
	case PLAYER.YOU :
	  moveStone(sp,gi) ;
	  if( msLbtn!=0 && (gi.mouseb & MK_LBUTTON)==0 ){
	    if( bd.setStone(put_x, put_y, your_stone)==0 ){
	      turn=(turn+1)%2 ;
	      playst=PLAYST.REVANIM ;
	    }
	  }
	  break ;
	}
      }else{
	ret=1 ;
      }
      break ;
    case PLAYST.REVANIM :
      playst=PLAYST.IDLE ;
      break ;
    }
    msLbtn = (gi.mouseb & MK_LBUTTON)!=0 ;
    setDrawPlayBD(sp,gi,DRAWST.IDLE) ;
    
    return(ret) ;
  }

  int moveStone(Sprite sp, GameInfo* gi){
    vec3D bd_nml ; bd_nml.set(0.0,1.0,0.0) ;
    bd_nml.rotate(cast(double)(g_rot_x)*PI*2.0/360.0, 1.0, 0.0, 0.0) ;
    vec3D bd_pos ; bd_pos.set(0.0,0.05,0.0) ;
    bd_pos.rotate(cast(double)(g_rot_x)*PI*2.0/360.0, 1.0, 0.0, 0.0) ;
    vec3D eyepos ; eyepos.set( sp.eyex,  sp.eyey,  sp.eyez) ;
    vec3D eyeray ;
    eyeray.set(( sp.eyez*tan(sp.fovy*PI/360.0)/(cast(double)(gi.windowHeight)/2.0) * (cast(double)(gi.mousex)-(cast(double)(gi.windowWidth )/2.0))) - sp.eyex, 
	       (-sp.eyez*tan(sp.fovy*PI/360.0)/(cast(double)(gi.windowHeight)/2.0) * (cast(double)(gi.mousey)-(cast(double)(gi.windowHeight)/2.0))) - sp.eyey,
		0.0-sp.eyez ) ;
    stonepos.LPcross(eyepos,eyeray,bd_pos,bd_nml) ;
    putpos=stonepos ;
    putpos.rotate(cast(double)(g_rot_x)*PI*2.0/360.0, -1.0, 0.0, 0.0) ;
    put_x = putpos.x+0.8<0 ? cast(int)((putpos.x+0.8-0.2)/0.2) : cast(int)((putpos.x+0.8)/0.2) ;
    put_y = putpos.z+0.8<0 ? cast(int)((putpos.z+0.8-0.2)/0.2) : cast(int)((putpos.z+0.8)/0.2) ;
    
    return(0) ;
  }

  int setDrawPlayBD(Sprite sp, GameInfo* gi, DRAWST drawst){
    Quaternion g_rot ;
    g_rot.setRotate(cast(double)(g_rot_x)*PI*2.0/360.0, 1.0, 0.0, 0.0) ;
    vec3D rotvec = g_rot.getAxis() ;
    
    sp.set3Draw(0,
		0.0,0.0,0.0,
		rotvec.x, rotvec.y, rotvec.z,
		g_rot.getAngle()*360/(PI*2.0),
		1.0, 1.0, 1.0, BLEND_TYPE.normal) ;
    
    for( int y=0 ; y<8 ; y++ ){
      for( int x=0 ; x<8 ; x++ ){
	vec3D pos ;
	pos.set(cast(double)(x-4)*0.2+0.1, 0.03, cast(double)(y-4)*0.2+0.1) ;
	pos.rotate(g_rot.getAngle(),g_rot.getAxis()) ;
	
	if( bd.matrix[y][x]!=STONE.NON ){
	  double rev = bd.matrix[y][x]==STONE.BLACK ? 180.0 : 0.0 ;
	  sp.set3Draw(1,
		      pos.x,pos.y,pos.z,
		      rotvec.x, rotvec.y, rotvec.z,
		      g_rot.getAngle()*360/(PI*2.0)+rev,
		      1.0, 1.0, 1.0, BLEND_TYPE.normal) ;
	}
      }
    }

    if( gamest==GAMEST.PLAYING && playst==PLAYST.PLAYER && player[turn]==PLAYER.YOU ){
      double rev = your_stone==STONE.BLACK ? 180.0 : 0.0 ;
      sp.set3Draw(1,
		  stonepos.x,stonepos.y,stonepos.z,
		  rotvec.x, rotvec.y, rotvec.z,
		  g_rot.getAngle()*360/(PI*2.0)+rev,
		  1.0, 1.0, 1.0, BLEND_TYPE.normal) ;
    }
    
    return(0) ;
  }

}
