import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.regexp;
import std.file;
import Vec3D;
import Object3D;

class OBJ3D
{
  BufferedFile infile ;
  Object3D     obj ;
  RegExp       getfstwd ;
  string       filepath ;
  string       curmaterial ;
  RegExp       obj_reg_tbl[string] ;
  RegExp       mtl_reg_tbl[string] ;
  int delegate(string, string) obj_act_tbl[string] ;
  int delegate(string, string) mtl_act_tbl[string] ;

  this(){
    string nume = "[-]*[0-9]+\\.[0-9]+[eE]{1}[-+]?[0-9]+" ;
    string numd = "[-]*[0-9]+\\.[0-9]+" ;
    string numi = "[-]*[0-9]+" ;
    string num  ;
    
    num = "(" ~ nume ~ "|" ~ numd ~ "|" ~ numi ~ ")" ;

    obj_reg_tbl["mtllib"] = new RegExp( "^mtllib\\s+(\\S+)"                            ,"") ;
    obj_reg_tbl["o"     ] = new RegExp( "^o\\s+(\\S+)"                                 ,"") ;
    obj_reg_tbl["s"     ] = new RegExp( "^s\\s+(\\S+)"                                 ,"") ;
    obj_reg_tbl["v"     ] = new RegExp( "^v\\s+"  ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    obj_reg_tbl["vn"    ] = new RegExp( "^vn\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    obj_reg_tbl["vt"    ] = new RegExp( "^vt\\s+" ~ num ~ "\\s+" ~ num                 ,"") ;
    obj_reg_tbl["g"     ] = new RegExp( "^g\\s+(\\S+)"                                 ,"") ;
    obj_reg_tbl["usemtl"] = new RegExp( "^usemtl\\s+(\\S+)"                            ,"") ;
    obj_reg_tbl["f"     ] = new RegExp( "^f\\s+([0-9]*)/([0-9]*)/([0-9]*)\\s+([0-9]*)/([0-9]*)/([0-9]*)\\s+([0-9]*)/([0-9]*)/([0-9]*)","") ;
    obj_reg_tbl["#"     ] = new RegExp( "^#"                                           ,"") ;

    obj_act_tbl["mtllib"] = &read_mtllib;
    obj_act_tbl["o"     ] = &alloc_obj  ;
    obj_act_tbl["v"     ] = &add_vertex ;
    obj_act_tbl["vn"    ] = &add_normal ;
    obj_act_tbl["vt"    ] = &add_texuv  ;
    //    obj_act_tbl["g"     ] = ;
    obj_act_tbl["usemtl"] = &chg_mtl    ;
    obj_act_tbl["f"     ] = &add_face   ;

    mtl_reg_tbl["newmtl"] = new RegExp( "^newmtl\\s+(\\S+)"                            ,"") ;
    mtl_reg_tbl["Ns"    ] = new RegExp( "^Ns\\s+" ~ num                                ,"") ;
    mtl_reg_tbl["d"     ] = new RegExp( "^d\\s+" ~ num                                 ,"") ;
    mtl_reg_tbl["illum" ] = new RegExp( "^illum\\s+" ~ num                             ,"") ;
    mtl_reg_tbl["Kd"    ] = new RegExp( "^Kd\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ka"    ] = new RegExp( "^Ka\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ks"    ] = new RegExp( "^Ks\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ke"    ] = new RegExp( "^Ke\\s+" ~ num ~ "\\s+" ~ num ~ "\\s+" ~ num  ,"") ;
    mtl_reg_tbl["map_Kd"] = new RegExp( "^map_Kd\\s+(\\S+)"                            ,"") ;
    
    mtl_act_tbl["newmtl"] = &add_mtl ;
    mtl_act_tbl["Ns"    ] = &add_mtl_Ns ;
    mtl_act_tbl["Kd"    ] = &add_mtl_Kd ;
    mtl_act_tbl["Ka"    ] = &add_mtl_Ka ;
    mtl_act_tbl["Ks"    ] = &add_mtl_Ks ;
    mtl_act_tbl["map_Kd"] = &add_mtl_mapKd ;

    curmaterial = "" ;
    
    getfstwd = new RegExp("^(\\S+)","") ;

  }

  ~this(){
    close() ;
    delete obj ;
  }
  
  int open(string fname, double scale=1.0){
    int i,len ;

    infile = new BufferedFile(fname) ;
    //writef("file=%s\n",fname) ;

    len=0 ;
    for( i=0 ; i<fname.length ; i++ ){
      if( fname[i]=='/' || fname[i]=='\\' ){
	len=i-1 ;
      }
    }
    if( len == 0 ){
      filepath = getcwd() ;
    }else{
      len++ ;
      filepath = getcwd() ~ '/' ~ fname[0..len] ;
    }
    filepath = std.array.replace(filepath, "\\", "/") ;

    obj = new Object3D(scale) ;
    obj.init() ;
    
    return(0) ;
  }

  int close(){
    if( !(infile is null) ){
      infile.close() ;
    }
    infile=null ;
    return(0) ;
  }

  int read() {
    string   line ;
    string   fstword ;
    string[] token ;
    string   word ;
    int i ;

    while(-1){
      line=cast(string)infile.readLine() ;
      if( line is null ) line="" ;
      if( line.length==0 ){
	if( infile.eof()==true ) break ;
	continue ;
      }
      
      //      writef("%d : %s\n",i++,line) ;
      token = getfstwd.split(line) ;

      if( token.length>=2 ){
	fstword = token[1].idup ;
	token=obj_reg_tbl["#"].match(fstword) ;
	if( token.length!=0 ) continue ;

	if( (fstword in obj_reg_tbl) == null ){
	  writef("keyword %s not found\n", fstword) ;
	  continue ;
	}
	if( (fstword in obj_act_tbl) != null ){
	  obj_act_tbl[fstword](fstword,line) ;
	}else{
	  //writef("internal error\n") ;
	  continue ;
	}
      }
    }

    obj.finish() ;
    return(0) ;
  }

  int read_mtllib(string act, string line){
    string       mtlline ;
    string       fstword ;
    string[]     token ;
    string       mtlfile ;
    BufferedFile inmtl ;

    //    writef("mtllib=%s\n",line) ;
    token = obj_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    
    mtlfile = filepath ~ '/' ~ token[1] ;
    inmtl = new BufferedFile(mtlfile) ;

    while(-1){
      mtlline=cast(string)inmtl.readLine() ;
      if( mtlline is null ) mtlline="" ;
      if( mtlline.length==0 ){
	if( inmtl.eof()==true ) break ;
	continue ;
      }
      token = getfstwd.split(mtlline) ;
      if( token.length>=2 ){
	fstword = token[1].idup ;
	token=obj_reg_tbl["#"].match(fstword) ;
	if( token.length!=0 ) continue ;
      
	if( (fstword in mtl_reg_tbl) == null ){
	  writef("Keyword %s not found\n", fstword) ;
	  continue ;
	}
	if( (fstword in mtl_act_tbl) != null ){
	  mtl_act_tbl[fstword](fstword,mtlline) ;
	}else{
	  //writef("internal error\n") ;
	  continue ;
	}
      }
    }
    inmtl.close() ;

    obj.finish_material() ;
    
    //    writef("read_mtllib\n") ;
    return(0) ;
  }

  int add_mtl(string act, string line){
    string[] token ;
    
    token = mtl_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_material(token[1]) ;
    return(0) ;
  }

  int add_mtl_Ns(string act, string line){
    string[] token ;
    
    token = mtl_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_mtl_Ns( std.conv.to!(double)(token[1]) ) ;
    return(0) ;
  }

  int add_mtl_Kd(string act, string line){
    string[] token ;
    
    token = mtl_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_mtl_Kd( std.conv.to!(double)(token[1]),std.conv.to!(double)(token[2]),std.conv.to!(double)(token[3]) ) ;
    return(0) ;
  }

  int add_mtl_Ka(string act, string line){
    string[] token ;
    
    token = mtl_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_mtl_Ka( std.conv.to!(double)(token[1]),std.conv.to!(double)(token[2]),std.conv.to!(double)(token[3]) ) ;
    return(0) ;
  }

  int add_mtl_Ks(string act, string line){
    string[] token ;
    
    token = mtl_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_mtl_Ks( std.conv.to!(double)(token[1]),std.conv.to!(double)(token[2]),std.conv.to!(double)(token[3]) ) ;
    return(0) ;
  }

  int add_mtl_mapKd(string act, string line){
    string[] token ;
    
    token = mtl_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_mtl_mapKd( filepath ~ '/' ~ token[1] ) ;
    return(0) ;
  }

  int alloc_obj(string act, string line){
    return(0) ;
  }
  
  int add_vertex(string act, string line){
    int      i ;
    string[] token ;
    
    token = obj_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_vertex( std.conv.to!(double)(token[1]),std.conv.to!(double)(token[2]),std.conv.to!(double)(token[3]) ) ;
    return(0) ;
  }

  int add_normal(string act, string line){
    int      i ;
    string[] token ;
    
    token = obj_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_normal( std.conv.to!(double)(token[1]),std.conv.to!(double)(token[2]),std.conv.to!(double)(token[3]) ) ;
    return(0) ;
  }

  int add_texuv(string act, string line){
    int      i ;
    string[] token ;
    
    token = obj_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    obj.add_texuv( std.conv.to!(double)(token[1]),std.conv.to!(double)(token[2]) ) ;
    return(0) ;
  }

  int chg_mtl(string act, string line){
    string[] token ;

    token = obj_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    curmaterial = token[1] ~ "" ;
    return(0) ;
  }

  int add_face(string act, string line){
    string[] token ;
    face3D   f,t,n ;
    
    token = obj_reg_tbl[act].split(line) ;
    if( token.length == 1 ){
      return(0) ;
    }
    
    if( token[1].length!=0 && token[4].length!=0 && token[7].length!=0 ){
      f.p0=std.conv.to!(int)(token[1]) ;
      f.p1=std.conv.to!(int)(token[4]) ;
      f.p2=std.conv.to!(int)(token[7]) ;
    }
    if( token[2].length!=0 && token[5].length!=0 && token[8].length!=0 ){
      t.p0=std.conv.to!(int)(token[2]) ;
      t.p1=std.conv.to!(int)(token[5]) ;
      t.p2=std.conv.to!(int)(token[8]) ;
    }
    if( token[3].length!=0 && token[6].length!=0 && token[9].length!=0 ){
      n.p0=std.conv.to!(int)(token[3]) ;
      n.p1=std.conv.to!(int)(token[6]) ;
      n.p2=std.conv.to!(int)(token[9]) ;
    }

    obj.add_face( curmaterial, f,t,n ) ;
    
    return(0) ;
  }
  
}

