import win32.windows;
import std.stdio;
import std.string;
import std.utf;
import win32frame.writek;
import win32frame.MsgTranslator;

shared static WindowBase[int] _WindowList ;
shared static WNDPROC[int]    _DefProc    ;

ushort LOWORD(uint l){ return cast(ushort)(l & 0xffff); }
ushort HIWORD(uint l){ return cast(ushort)(l >>> 16)  ; }

class WindowBase {
  HWND        hwnd;
  WNDCLASSEX  winc;
  string      name ;
  HMENU       hmenu ;
  bool        finish_create ;
  
  this( string name ){
    this.winc.cbSize      = WNDCLASSEX.sizeof;
    this.winc.hInstance   = GetModuleHandle(null) ;
    this.winc.lpfnWndProc = cast(WNDPROC)(&_WndProc) ;
    this.hwnd             = null ;
    this.name             = name ;
    this.finish_create    = false;
  }

  int setWNDCLASSEX( WNDCLASSEX winc){
    this.winc=winc ;
    return(0) ;
  }

  int registwinc(){
    if( !RegisterClassEx(&this.winc) ) return(-1) ;
    return(0) ;
  }
  
  int lowcreate(DWORD exstyle, DWORD style,
		int x, int y, int width, int height, HWND phwnd, HMENU hmenu){
    
    if( this.finish_create==false ){
      HWND rethwnd ;
      this.hwnd=null ;
      rethwnd = CreateWindowEx(exstyle,
			       this.winc.lpszClassName,
			       cast(wchar*)(toUTF16z(this.name)),
			       style,
			       x,y,width,height,
			       phwnd, hmenu, this.winc.hInstance , cast(LPVOID)(this)
			       );
      if( rethwnd == null ){
	writek("create failed %s %d,%d,%d,%d\n",name, x,y,width,height) ;
	return(-2) ;
      }else{
	this.hwnd=rethwnd ;
      }
      _WindowList[cast(int)(this.hwnd)]=cast(shared WindowBase)(this) ;
      _DefProc[cast(int)(this.hwnd)]   =cast(shared WNDPROC)(GetWindowLong(this.hwnd , GWL_WNDPROC)) ;
      SetWindowLong( this.hwnd, GWL_WNDPROC, cast(int)(&_WndSubProc)) ;
      this.finish_create=true ;
      this.hmenu=hmenu ;
      writek("create hwnd=%08x name=%s\n",cast(int)(this.hwnd),name) ;
      return(0) ;
    }
    return(-1) ;
  }

  HWND getHWND(){
    return( this.hwnd ) ;
  }
  
  HWND getParentHWND(){
    return( GetParent(this.hwnd) ) ;
  }
  
  HINSTANCE getInstance(){
    return( this.winc.hInstance ) ;
  }

  HMENU getHMENU(){
    return( this.hmenu ) ;
  }

  int setFont(HFONT hfont){
    if( this.hwnd!=null ){
      SendMessage( this.hwnd, WM_SETFONT, cast(WPARAM)(hfont), cast(LPARAM)(TRUE) );
      return(0) ;
    }
    return(-1) ;
  }

  LRESULT sendmsg(UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(this.hwnd,msg,wp,lp)) ;
  }

  LRESULT sendmsg(HWND tgthwnd, UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(tgthwnd,msg,wp,lp)) ;
  }

  LRESULT sendmsgtoparent(UINT msg , WPARAM wp , LPARAM lp){
    return(SendMessage(this.getParentHWND(),msg,wp,lp)) ;
  }

  BOOL setWindowText(string str){
    int res = SendMessage( this.hwnd, WM_SETTEXT, 0, cast(LPARAM)(cast(wchar*)(toUTF16z(str))) ) ;
    if( res<=0 ){
      return(false) ;
    }
    return(true) ;
  }

  string getWindowTextUTF16(){
    char[] str = new char[GetWindowTextLength(this.hwnd)*2+2] ;
    GetWindowText(this.hwnd , cast(wchar*)(str) , str.length);
    str.length=str.length-2 ;
    return(str.idup) ;
  }
  
  string getWindowText(){
    char[]  str  = new char[GetWindowTextLength(this.hwnd)*2+2] ;
    wchar[] wstr   ;
    string  retstr ;
    int     len=0  ;
    GetWindowText(this.hwnd , cast(wchar*)(str) , str.length);
    str.length =str.length-2 ;
    wstr.length=str.length ;
    for( int i=0 ; i<str.length ; i+=2 ){
      wstr[i/2] = (cast(ushort)(str[i+1])<<8 | cast(ushort)(str[i])) ;
      len++ ;
    }
    wstr.length=len ;
    retstr = std.utf.toUTF8(wstr.idup) ;
    return(retstr) ;
  }
  
  LRESULT preproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return( this.proc(hwnd , msg , wp , lp) ) ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return( DefWindowProc(hwnd , msg , wp , lp) ) ;
  }

  LRESULT subproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return(-1) ;
  }

  LRESULT presubproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    return(-1) ;
  }

  ~this(){
    writek("----WindowBase called destructor hwnd=%08x\n",this.hwnd) ;
    if( this.hwnd!=null ){
      DestroyWindow(this.hwnd);
    }
  }
}

extern (Windows) LRESULT _WndProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  version(ENBMSGTRACE){
    writek("_WndProc : hwnd=%08x, msg=%s\n",cast(int)(hwnd),MsgTranslate(msg)) ;
  }
  if( msg==WM_NCCREATE ){
    synchronized{
      WindowBase wb = cast(WindowBase)(*(cast(int*)(lp))) ;
      wb.hwnd=hwnd ;
      _WindowList[cast(int)(hwnd)]=cast(shared WindowBase)(wb) ;
    }
  }
  
  if( (cast(int)(hwnd) in _WindowList) != null ){
    WindowBase wb=cast(WindowBase)(_WindowList[cast(int)(hwnd)]) ;
    LRESULT   res=wb.preproc(hwnd,msg,wp,lp) ;
    if( msg==WM_NCDESTROY ){
      synchronized{
	_WindowList[cast(int)(hwnd)]=null ;
	_WindowList.remove(cast(int)(hwnd)) ;
	_DefProc[cast(int)(hwnd)]=null ;
	_DefProc.remove(cast(int)(hwnd)) ;
	wb.hwnd=null ;
      }
    }
    return( res ) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}

extern (Windows) LRESULT _WndSubProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp)
{
  version(ENBMSGTRACE){
    writek("_WndSubProc : hwnd=%08x, msg=%s\n",cast(int)(hwnd),MsgTranslate(msg)) ;
  }
  if( (cast(int)(hwnd) in _DefProc) != null ){
    WindowBase wb=cast(WindowBase)(_WindowList[cast(int)(hwnd)]) ;
    wb.presubproc(hwnd,msg,wp,lp) ;
    WNDPROC wproc = cast(WNDPROC)(_DefProc[cast(int)(hwnd)]) ;
    return(CallWindowProc( wproc , hwnd, msg, wp, lp)) ;
  }
  return( DefWindowProc(hwnd , msg , wp , lp) ) ;
}
